/***************************************************************************
 * First Cut at Ports0 interface for my adaptation of new threads 
 * Author: Neel Sundaresan, Indiana University
 ***************************************************************************/

#ifndef _NewThreads_h_
#define _NewThreads_h_ 1

/*****************************************************************************
 *
 *  first, define some system-dependent macros 
 */

/* Unless told otherwise, assume one-node configuration */
#define OneNode 

/* These entries are system definitions -- for backwards compatibility. */

#ifdef _SGI
#define Mips
#else
#ifdef _PARAGON
#define Ipsc860
#define STANDALONE
#else
asdjfaldjflfj
#endif
#endif

#if defined(Mips)
#define StackGrowsUp    0
#endif

#ifdef Ipsc860
#define I860      1
#ifndef Use_IntelNX
#define Use_IntelNX
#endif
#endif


/* The following are Architecture definitions */

#ifdef I860
#define	StackGrowsUp	0	
#endif

#ifdef Use_IntelNX
#ifdef OneNode
#undef OneNode
#endif
#define	IntelNX
#endif



/****************************************************************************/
/*  Message related stuff */

#define LowMsg (0)
#define HighMsg (63)
#define NumMsgPorts (HighMsg-LowMsg+1)

#define SyncLowMsg (LowMsg)
#define SyncHighMsg (HighMsg/2)

#define SysSyncLowMsg (SyncLowMsg)
#define SysSyncHighMsg ((SyncLowMsg+SyncHighMsg)/2)

#define UserSyncLowMsg (SysSyncHighMsg+1)
#define UserSyncHighMsg (SyncHighMsg)

#define AsyncLowMsg (SyncHighMsg+1)
#define AsyncHighMsg (HighMsg)

#define SysAsyncLowMsg (AsyncLowMsg)
#define SysAsyncHighMsg ((AsyncLowMsg+AsyncHighMsg)/2)
 
#define UserAsyncLowMsg (SysAsyncHighMsg+1)
#define UserAsyncHighMsg (AsyncHighMsg)

/* some messages which are used at C-level message passing */
#define C_BarrierFanInMsg (SysSyncLowMsg+0)
#define C_BarrierFanOutMsg (SysSyncLowMsg+1)

#define C_SyncMsgs 2
#define C_AsyncMsgs 0

/****************************************************************************/
/* system include files and defines for different targets */
/****************************************************************************/
#ifdef _SGI
/* trying out 32 byte/64byte shared locks */
#define _SHARE_32_64 0

/* which shared lock to use: */
#define SHMUTEX1 1

#include <ulocks.h>
/* prctl related stuff  */
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/prctl.h>

#if _SHARE_32_64
#include <sgidefs.h>
#include <abi_mutex.h>
#endif
/* m_fork kind of calls */
#include <ulocks.h>
#include <task.h>
#include <assert.h>
#include <unistd.h>
#endif /* _SGI */

typedef char Boolean;
#ifdef TRUE
#undef TRUE
#endif
#ifdef FALSE
#undef FALSE
#endif
#define TRUE 1
#define FALSE 0

/****************************************************************************/
/* various options for compile time settings and optimizations based */
/* on the target  environment */

#if 0
#define _COMMUNICATION_ON /* this should be coupled with other msg */
			  /* passing options */ 
#endif
#define _GRACEFUL_EXIT /* graceful exit any aborted gracefully kills others */

#if 0
#ifdef _STATISTICS_ON /* only if thread statistics is on it makes */
		       /* sense  to have the timer on */
#define _TIMER_ON         /* is ports0 timer turned on */
#endif
#define _SWITCH_SAFE      /* minimal work done at the time of switch */
#define _JOIN_BY_COND /* thread join should be done using condition or yield */
#endif

#ifdef _PARAGON
#define _SINGLE_PROCESSOR_PER_PC /* only one processor per pc */
#define _SINGLE_PC_PER_SD        /* single pc for every sd */
#define _SINGLE_DOMAIN           /* single domain */
#endif


#ifdef _SGI
#define _SINGLE_PROCESSOR_PER_PC /* only one processor per pc */
#define _SINGLE_PC_PER_SD        /* single pc for every sd */
#if 0
#endif
#define _SINGLE_SD_PER_D         /* only one sd in the domain */
#define _SINGLE_DOMAIN           /* single domain */
#endif
/****************************************************************************/


/****************************************************************************/
#ifdef _TIMER_ON
#include "../ports/ports_timers.h"
#define ReadyTimerId(t) (ThreadGetRegId(t))
#define BlockedTimerId(t) (20+ThreadGetRegId(t))
#define LifeTimerId(t) (40+ThreadGetRegId(t))
#define LockTimerId(t) (63-ThreadGetRegId(t))
#endif
/****************************************************************************/


#ifdef _SGI
#define MAX_PROCESSORS 16
#define MAX_CONTEXTS MAX_PROCESSORS
typedef usptr_t* MemoryHandle;
#endif

#ifdef _PARAGON
#define MAX_PROCESSORS 1
#define MAX_CONTEXTS 1
typedef void* MemoryHandle;
#endif

/****************************************************************************/
/*
 * spinlocks: Process level
 */
/****************************************************************************/

#ifdef _SGI
struct _SpinlockStruct
{
   ulock_t lock;
   MemoryHandle mem;
   int locker;         /* thread which owns the lock */
   int processor;      /* processor on which the thread is sitting */
};
typedef struct _SpinlockStruct Spinlock;
#else

#ifdef _PARAGON
typedef int Spinlock;
#else

 what is the name of this machine?
#endif
#endif

void InitSpinlock(MemoryHandle, Spinlock**); 

void DestroySpinlock(Spinlock*);

void AcquireSpinlock(Spinlock*);

void ReleaseSpinlock(Spinlock*);


/* Spinlock between processes at the context and the subdomain level */
typedef Spinlock PCSpinlock;
typedef Spinlock SDSpinlock;

#ifdef _SINGLE_PROCESSOR_PER_PC
#define InitPCSpinlock(mem, lockp) /* no-op */
#define DestroyPCSpinlock(lock) /* no-op */
#define AcquirePCSpinlock(lock) /* no-op */
#define ReleasePCSpinlock(lock) /* no-op */
#else
#define InitPCSpinlock(mem, lockp) InitSpinlock(mem, lockp)
#define DestroyPCSpinlock(lock) DestroySpinlock(lock)
#define AcquirePCSpinlock(lock) AcquireSpinlock(lock)
#define ReleasePCSpinlock(lock) ReleaseSpinlock(lock)
#endif

#ifdef _SINGLE_PC_PER_SD
#define InitSDSpinlock(mem, lockp) InitPCSpinlock(mem, lockp)
#define DestroySDSpinlock(lock) DestroyPCSpinlock(lock)
#define AcquireSDSpinlock(lock) AcquirePCSpinlock(lock)
#define ReleaseSDSpinlock(lock) ReleasePCSpinlock(lock)
#else
#define InitSDSpinlock(mem, lockp) InitSpinlock(mem, lockp)
#define DestroySDSpinlock(lock) DestroySpinlock(lock)
#define AcquireSDSpinlock(lock) AcquireSpinlock(lock)
#define ReleaseSDSpinlock(lock) ReleaseSpinlock(lock)
#endif

/****************************************************************************/
/* Mutexes:
 * shared_Mutexes: added by neel
 * these mutexes are used to share locks between threads across
 * processor domains.   
 *
 * Mutexes are (blocking) locks.  Only one thread at a time may hold a mutex;
 * additional threads trying to acquire the mutex will block until the mutex
 * is free.  
 *
 * A mutex is acquired by calling mutexAcquire, and released by calling
 * mutexRelease.
 *
 * Mutexes must be initialized before use (by calling mutexInit).  They
 * should be destroyed after use (by calling mutexDestroy).
 *
 * Mutexes are not FIFO; a thread trying to acquire the mutex may be infinitely
 * overtaken by other threads acquiring that mutex.  (Fairness has a 
 * performance cost, so we decided to do without it.)
 *
 * The above is the old performance (obtainable with OLD_MUTEX_BEHAVIOUR)
 * The current default is to have mutexs queue requests.
 */
/****************************************************************************/

typedef struct 
{
  int              locked;
  struct Thread*   locker;
  int              recurse_count;
  struct Thread*   waiters;
  MemoryHandle     mem;
} LocMutex;

void local_mutexInit(MemoryHandle, LocMutex**);
void local_mutexDestroy(LocMutex*);
void local_mutexAcquire(LocMutex*);
void local_mutexRelease(LocMutex*);


#ifdef _SGI
#if SHMUTEX1
struct _ShMutex
{
   MemoryHandle    mem_handle; /* memory from which the fields are allocated */
#if _SHARE_32_64
   abilock_t*        us_lock;
#else
   ulock_t         us_lock;
#endif
   LocMutex**      mutexes; 
   struct Thread** someone_in; /* array of thread_ptrs for each pc  */
   int             processor;  /* processor on which the owning thread is */
/* the following 2 fields are for monitoring purposes only! */
   int             local_count;
   int             shared_count;
};
typedef struct _ShMutex  ShMutex;
#endif
#else
#ifdef _PARAGON
typedef LocMutex ShMutex;
#else
 not implemented!
#endif
#endif

#if SHMUTEX1
void shared_mutexInit(MemoryHandle, ShMutex**);
void shared_mutexDestroy(ShMutex*);
void shared_mutexAcquire(ShMutex*);
void shared_mutexRelease(ShMutex*);
#endif


/* the following is defined to compare it with our implementation of */
/* shared_mutex  and to do timing tests.. not used otherwise         */
#ifdef _SGI
typedef Spinlock ShMutex2;
#if !SHMUTEX1
typedef  ShMutex2 ShMutex;
#endif
void shared2_mutexInit(MemoryHandle, ShMutex2**);
void shared2_mutexDestroy(ShMutex2*);
void shared2_mutexAcquire(ShMutex2*);
void shared2_mutexRelease(ShMutex2*);
#if !SHMUTEX1
#define shared_mutexInit(x,y) shared2_mutexInit(x,y)
#define shared_mutexDestroy(x) shared2_mutexDestroy(x)
#define shared_mutexAcquire(x) shared2_mutexAcquire(x)
#define shared_mutexRelease(x) shared2_mutexRelease(x)
#endif
#endif

#ifdef _SINGLE_PROCESSOR_PER_PC
typedef LocMutex PCMutex;
#define PCMutexInit(mem, mutexp_p) local_mutexInit(mem, mutexp_p)
#define PCMutexDestroy(mutexp) local_mutexDestroy(mutexp)
#define PCMutexAcquire(mutexp) local_mutexAcquire(mutexp)
#define PCMutexRelease(mutexp) local_mutexRelease(mutexp)
#else
typedef ShMutex PCMutex; 
#define PCMutexInit(mem, mutexp_p) shared_mutexInit(mem, mutexp_p)
#define PCMutexDestroy(mutexp) shared_mutexDestroy(mutexp)
#define PCMutexAcquire(mutexp) shared_mutexAcquire(mutexp)
#define PCMutexRelease(mutexp) shared_mutexRelease(mutexp)
#endif

#ifdef _SINGLE_PC_PER_SD
typedef int SDShallowMutex;
#define SDShallowMutexInit(mem, mutexp_p) /*no-op*/
#define SDShallowMutexDestroy(mutexp)  /*no-op*/
#define SDShallowMutexAcquire(mutexp) /*no-op*/
#define SDShallowMutexRelease(mutexp)  /*no-op*/
#else
/* we have not yet defined SD level locks so we will use ProcessLocks */
/* from above. */
typedef SDSpinlock SDShallowMutex;
#define SDShallowMutexInit(mem, mutexp_p) InitSDSpinlock(mem, mutexp_p)
#define SDShallowMutexDestroy(mutexp)  DestroySDSpinlock(mutexp)
#define SDShallowMutexAcquire(mutexp) AcquireSDSpinlock(mutexp)
#define SDShallowMutexRelease(mutexp)  ReleaseSDSpinlock(mutexp)
#endif

#ifdef _SINGLE_PC_PER_SD
typedef PCMutex SDDeepMutex;
#define SDDeepMutexInit(mem, mutexp_p) PCMutexInit(mem, mutexp_p)
#define SDDeepMutexDestroy(mutexp)  PCMutexDestroy(mutexp)
#define SDDeepMutexAcquire(mutexp) PCMutexAcquire(mutexp)
#define SDDeepMutexRelease(mutexp)  PCMutexRelease(mutexp)
#else
/* we have not done this part... and whatever we have done, is wrong!!! */
typedef SDSpinlock SDDeepMutex;
#define SDDeepMutexInit(mem, mutexp_p) assert(0)
#define SDDeepMutexDestroy(mutexp)  assert(0)
#define SDDeepMutexAcquire(mutexp) assert(0)
#define SDDeepMutexRelease(mutexp)  assert(0)
#endif

/****************************************************************************/
/* Conditions are condition-variables with Mesa semantics.  (If you don't 
 * understand what that means, never mind.)  
 * 
 * Conditions must be initialized before use (by calling conditionInit).  They
 * should be destroyed after use (by calling conditionDestroy).
 *
 * Conditions are not FIFO; a thread waiting on a condition may be infinitely
 * overtaken by other threads waiting on that condition.  (Fairness has a 
 * performance cost, so we decided to do without it.)
 *
 * The above is the old performance (obtainable with OLD_CONDITION_BEHAVIOUR)
 * The current default is to have conditions queue requests.
 */
/****************************************************************************/

typedef struct 
{
   struct Thread* waiters;
   MemoryHandle mem;
   ShMutex* cond_lock;
} ShCondition;

void shared_conditionInit(MemoryHandle, ShCondition**);
void shared_conditionDestroy(ShCondition*);
void shared_conditionWait(ShCondition*, ShMutex*);
void shared_conditionSignal(ShCondition*);
void shared_conditionBroadcast(ShCondition*);

/* LOCAL condition stuff */
typedef struct _LocCondition
{
   struct Thread* waiters;
   MemoryHandle mem;
} LocCondition;

void local_conditionInit(MemoryHandle, LocCondition**);
void local_conditionDestroy(LocCondition*);
void local_conditionWait(LocCondition*, LocMutex*);
void local_conditionSignal(LocCondition*);
void local_conditionBroadcast(LocCondition*);

#ifdef _SINGLE_PROCESSOR_PER_PC
typedef LocCondition PCCondition;
#define PCConditionInit(mem, condp_p) local_conditionInit(mem, condp_p)
#define PCConditionDestroy(condp)     local_conditionDestroy(condp)
#define PCConditionWait(condp, mutexp) local_conditionWait(condp, mutexp)
#define PCConditionSignal(condp)      local_conditionSignal(condp)
#define PCConditionBroadcast(condp) local_conditionBroadcast(condp)
#else
typedef ShCondition PCCondition;
#define PCConditionInit(mem, condp_p) shared_conditionInit(mem, condp_p)
#define PCConditionDestroy(condp)     shared_conditionDestroy(condp)
#define PCConditionWait(condp, mutexp) shared_conditionWait(condp, mutexp)
#define PCConditionSignal(condp)      shared_conditionSignal(condp)
#define PCConditionBroadcast(condp) shared_conditionBroadcast(condp)
#endif

#ifdef _SINGLE_PC_PER_SD
typedef int SDShallowCondition;
#define SDShallowConditionInit(mem, condp_p)         /* no-op */
#define SDShallowConditionDestroy(condp)             /* no-op */
#define SDShallowConditionWait(condp, mutexp)        /* no-op */
#define SDShallowConditionSignal(condp)              /* no-op */
#define SDShallowConditionBroadcast(condp)           /* no-op */
#else
/* the following is not defined yet! whatever is, is wrong */
typedef ShCondition SDShallowCondition;
#define SDShallowConditionInit(mem, condp_p)  assert(0)
#define SDShallowConditionDestroy(condp)      assert(0)
#define SDShallowConditionWait(condp, mutexp) assert(0)
#define SDShallowConditionSignal(condp)       assert(0)
#define SDShallowConditionBroadcast(condp)    assert(0) 
#endif

#ifdef _SINGLE_PC_PER_SD
typedef PCCondition SDDeepCondition;
#define SDDeepConditionInit(mem, condp_p) PCConditionInit(mem, condp_p)
#define SDDeepConditionDestroy(condp)     PCConditionDestroy(condp)
#define SDDeepConditionWait(condp, mutexp) PCConditionWait(condp, mutexp)
#define SDDeepConditionSignal(condp)      PCConditionSignal(condp)
#define SDDeepConditionBroadcast(condp)   PCConditionBroadcast(condp)
#else
/* the following is not defined yet! whatever is, is wrong */
typedef ShCondition SDDeepCondition;
#define SDDeepConditionInit(mem, condp_p)  assert(0)
#define SDDeepConditionDestroy(condp)      assert(0)
#define SDDeepConditionWait(condp, mutexp) assert(0)
#define SDDeepConditionSignal(condp)       assert(0)
#define SDDeepConditionBroadcast(condp)    assert(0) 
#endif

/****************************************************************************/
/* Barriers:
 * Process level barriers.
 */
/****************************************************************************/

#ifdef _SGI
typedef barrier_t* ProcBarrier;
#else
#ifdef _PARAGON
typedef int ProcBarrier;
#endif
#endif

typedef ProcBarrier PCProcBarrier;
typedef ProcBarrier SDProcBarrier;

void InitProcBarrier(MemoryHandle mem, ProcBarrier* barrier_p);
void SynchProcBarrier(const ProcBarrier b, const int count);
void DestroyProcBarrier(const ProcBarrier b);

#ifdef _SINGLE_PROCESSOR_PER_PC
#define InitPCProcBarrier(mem, barrier_p) /* no-op */
#define SynchPCProcBarrier(barrier, count) /* no-op */
#define DestroyPCProcBarrier(barrier) /* no-op */
#else
#define InitPCProcBarrier(mem, barrier_p) InitProcBarrier(mem, barrier_p)
#define SynchPCProcBarrier(barrier,count) SynchProcBarrier(barrier,(count))
#define DestroyPCProcBarrier(barrier) DestroyProcBarrier(barrier)
#endif

#ifdef _SINGLE_PC_PER_SD
#define InitSDProcBarrier(mem, barrier_p) InitPCProcBarrier(mem, barrier_p)
#define SynchSDProcBarrier(barrier, count) SynchPCProcBarrier(barrier, count)
#define DestroySDProcBarrier(barrier) DestroyPCProcBarrier(barrier)
#else
#define InitSDProcBarrier(mem, barrier_p) InitProcBarrier(mem, barrier_p)
#define SynchSDProcBarrier(barrier,count) SynchProcBarrier(barrier,(count))
#define DestroySDProcBarrier(barrier)      DestroyProcBarrier(barrier)
#endif

#ifdef _SINGLE_SD_PER_D
#define InitDomProcBarrier(mem, barrier_p) InitSDProcBarrier(mem, barrier_p)
#define SynchDomProcBarrier(barrier, count) SynchSDProcBarrier(barrier,(count))
#else
#define InitDomProcBarrier(mem, barrier_p) assert(0)
#define SynchDomProcBarrier(barrier, count) assert(0)
#endif

/****************************************************************************/
/*
 * The following are thread-level barriers
 */
/****************************************************************************/
typedef struct LocBarrier_struct
{
   LocMutex* gate;  
   LocCondition* in_latch;
   LocCondition* out_latch;
   LocCondition* in_extra_latch;
   LocCondition* out_extra_latch;
   int in_barrier;
   int out_barrier;
   char* in_check_list;
   char* out_check_list;
   int size;
   long in_counter;
} LocThrBarrier;

void local_barrierInit(MemoryHandle mem, 
		       LocThrBarrier** barrier_ptr_p,
		       const int barrier_size);
void local_barrierDestroy(LocThrBarrier* barrier_ptr);
int local_barrierCheckin(LocThrBarrier*, const int);
int local_barrierCheckout(LocThrBarrier*, const int);
void local_barrierSynch(LocThrBarrier* barrier_ptr, const int my_id);


/* thread barrier: checkin/checkout difficult */
typedef struct
{
  struct _treenode* nodes;
  int      size;
  MemoryHandle mem;
  Boolean sense;
} ShThrBarrier;  

void shared_barrierInit(MemoryHandle mem, 
			ShThrBarrier** barrier_ptr_p,
			const int barrier_size); 
void shared_barrierDestroy(ShThrBarrier* barrier_ptr);
void shared_barrierSynch(ShThrBarrier* barrier_ptr, 
			 const int my_id);
/* fuzzy barrier stuff */
void shared_barrierCheckin(ShThrBarrier*, const int);
void shared_barrierCheckout(ShThrBarrier*, const int);

#ifdef _SINGLE_PROCESSOR_PER_PC
typedef LocThrBarrier PCThrBarrier;
#define PCThrBarrierInit(mem, bp, barrier_size) local_barrierInit(mem,\
								  bp,\
							      barrier_size) 
#define PCThrBarrierDestroy(barrier) local_barrierDestroy(barrier)
#define PCThrBarrierSynch(barrier, my_id) local_barrierSynch(barrier,\
							     my_id)
#define PCThrBarrierCheckin(barrier, my_id) local_barrierCheckin(barrier,\
								 my_id)
#define PCThrBarrierCheckout(barrier, my_id) local_barrierCheckout(barrier,\
								 my_id)
#else
typedef ShThrBarrier PCThrBarrier;
#define PCThrBarrierInit(mem, bp, barrier_size) shared_barrierInit(mem,\
						       bp, barrier_size) 
#define PCThrBarrierDestroy(barrier) shared_barrierDestroy(barrier)
#define PCThrBarrierSynch(barrier, my_id) shared_barrierSynch(barrier,\
							     my_id)
#define PCThrBarrierCheckin(barrier, my_id) shared_barrierCheckin(barrier,\
								 my_id)
#define PCThrBarrierCheckout(barrier, my_id)\
                                 shared_barrierCheckout(barrier, my_id)

#endif

#ifdef _SINGLE_PC_PER_SD
typedef  int SDShallowThrBarrier;
#define SDShallowThrBarrierInit(mem, bp, barrier_size) /* no-op */
#define SDShallowThrBarrierDestroy(barrier)    /* no-op */
#define SDShallowThrBarrierSynch(barrier, my_id) /* no-op */
#define SDShallowThrBarrierCheckin(barrier, my_id) /* no-op */
#define SDShallowThrBarrierCheckout(barrier, my_id) /* no-op */
#else
/* only thread per PC is trying to synchronize. Is this good enuf? */
typedef ShThrBarrier SDShallowThrBarrier;
#define SDShallowThrBarrierInit(mem, bp, barrier_size) shared_barrierInit(mem,\
						       bp, barrier_size) 
#define SDShallowThrBarrierDestroy(barrier) shared_barrierDestroy(barrier)
#define SDShallowThrBarrierSynch(barrier, my_id) shared_barrierSynch(barrier,\
							     my_id)
#define SDShallowThrBarrierCheckin(barrier, my_id) shared_barrierCheckin(barrier,\
								 my_id)
#define SDShallowThrBarrierCheckout(barrier, my_id)\
                                 shared_barrierCheckout(barrier, my_id)
#endif

#if 0 /* we are not sure if we need SDDeepThrBarrier */
#ifdef _SINGLE_PC_PER_SD
typedef PCThrBarrier SDDeepThrBarrier;
#define SDDeepThrBarrierInit(mem, bp, barrier_size) PCThrBarrierInit(mem,\
						    bp, barrier_size) 
#define SDDeepThrBarrierDestroy(barrier) PCThrBarrierDestroy(barrier)
#define SDDeepThrBarrierSynch(barrier, my_id) PCThrBarrierSynch(barrier,\
							     my_id)
#define SDDeepThrBarrierCheckin(barrier, my_id) PCThrBarrierCheckin(barrier,\
								 my_id)
#define SDDeepThrBarrierCheckout(barrier, my_id) PCThrBarrierCheckout(barrier,\
								 my_id)
#else
typedef ShThrBarrier SDDeepThrBarrier;
#define SDDeepThrBarrierInit(mem, bp, barrier_size) shared_barrierInit(mem,\
								bp,\
							      barrier_size) 
#define SDDeepThrBarrierDestroy(barrier) shared_barrierDestroy(barrier)
#define SDDeepThrBarrierSynch(barrier, my_id) shared_barrierSynch(barrier,\
							     my_id)
#define SDDeepThrBarrierCheckin(barrier, my_id) shared_barrierCheckin(barrier,\
								 my_id)
#define SDDeepThrBarrierCheckout(barrier, my_id) shared_barrierCheckout(barrier,\
								 my_id)
#endif

#endif /* */

/* the following needs to be fixed! */
typedef struct
{
  int thread_id;
  int context;
  int node;
} ThreadAddrStruct;
typedef struct
{
  int size;
  ThreadAddrStruct** thr_addr_list;
  int i_have_reached_the_barrier;
  MemoryHandle mem;
} DistThrBarrier;

void dist_barrierInit(MemoryHandle mem, 
		      DistThrBarrier** barrier_ptr_p,
		      ThreadAddrStruct** t_attr_list,
		      const int barrier_size);
void dist_barrierDestroy(DistThrBarrier* barrier_ptr);
int dist_barrierCheckin(DistThrBarrier*, const int);
int dist_barrierCheckout(DistThrBarrier*, const int);
void dist_barrierSynch(DistThrBarrier* barrier_ptr);

#ifdef _SINGLE_SD_PER_D
typedef  int DomShallowThrBarrier;
#define DomShallowThrBarrierInit(mem, bp, thr_addr_list, barrier_size) /* no-op */
#define DomShallowThrBarrierDestroy(barrier_p) /* no-op */
#define DomShallowThrBarrierSynch(barrier_p,me) /* no-op */
#define DomShallowThrBarrierCheckin(barrier_p,me) /* no-op */
#define DomShallowThrBarrierCheckout(barrier_p,me) /* no-op */
#else
typedef DistThrBarrier DomShallowThrBarrier;
#define DomShallowThrBarrierInit(mem, bp, thr_list, barrier_size) \
                              dist_barrierInit(mem, bp, thr_list, barrier_size)
#define DomShallowThrBarrierDestroy(barrier_p) \
                              dist_barrierDestroy(barrier_p)
#define DomShallowThrBarrierSynch(barrier_p, my_id) \
                              dist_barrierSynch(barrier_p)
#define DomShallowThrBarrierCheckin(barrier_p,my_id) \
                              dist_barrierCheckin(barrier_p, my_id)
#define DomShallowThrBarrierCheckout(barrier_p,my_id) \
                              dist_barrierCheckout(barrier_p, my_id)
#endif

/* domain  mid thread barrier */

/* domain deep thread barrier */
  
/*****************************************************************************
 *
 * threads and synchronization 
 */


#define	ThreadStackSize	  (64*1024)

typedef enum { 
  Defunct,   /* uninitialized or dead */
  Virgin,    /* ready, but never been run */
  Ready,     /* ready to run */
  Running,   /* currently running on a CPU */
  Blocked   /* waiting for something to happen */
} ThreadStatus;


typedef void (*ThreadFunction)(void*);  /* type of a thread's start-function */

typedef int thread_key_t; /* neel */
typedef long thread_data_t; /* neel */
#define MAX_PVT_DATA 8              /* neel */


/***************************************************************************
  *
  * Communication related stuff - moved here from comm.c 
  *
 ***************************************************************************/


typedef struct MsgInfo {
   int sender_node;
   int sender_pc; /* should this be here or at c++ level? */
   int sender_thread_id;
   int msg_type; 
} MsgInfo;

typedef struct BufferDescriptor {
  void* buf;
  int   size;
  MsgInfo info_struct;
  struct BufferDescriptor* next;
} BufferDescriptor;

typedef struct {
  struct Thread*  waiters; /* chained by comm_next field in the thread */
  PCMutex* lock;
} CommWaiters;

typedef struct {
  PCMutex* lock;
  
  BufferDescriptor* firstBuf;   /* waiting buffers */
  BufferDescriptor* lastBuf;
} CommPort;


#ifndef _SINGLE_PC_PER_SD
/* for now we define the CommPort to be a MailBox we will not use the waiters
   field at all */
typedef struct MailEntry {
  void* buf;
  int   size;
  MsgInfo info_struct;
  int for_thread;
  struct MailEntry* next;
} MailEntry;

typedef struct {
  SDSpinlock* lock;
  MemoryHandle mem;
  MailEntry* first;   /* waiting buffers */
  MailEntry* last;   /* waiting buffers */

} MailBox;
#endif

/****************** end of stuff moved from comm.c **************************/


typedef struct Thread {
  void*          userStuff;       /* user can do what s/he wants with this */

  void*          saveSp;          /* saved stack-pointer */

  ThreadStatus   status;          

  ThreadFunction startFunc;       /* function that thread should execute */
  void*          startArg;        /* argument for that function */

  struct Thread* readyQ;          /* used for linking the ready-thread queue */
  struct Thread* mutex_next;      /* used for locking purposes */
  struct Thread* comm_next;       /* used for communication purposes */

  void*          msgBuf;         /* describe buffer that thread is waiting */ 

  int            msgSize;        /*   to receive into */

  int            usesFloatingPoint; /* does this thread use floating-point? */

  char           name[8];          /* for debugging only */

  thread_data_t  pdata[MAX_PVT_DATA]; /* there can be only 8 pvt fields: Neel */
  int            num_pvt_data; /* number of pvt datas: Neel */
  double         dummy; /* for double boundary algnment of stack */
#if 0
  char           stack[ThreadStackSize];  /* the thread's stack */
#else
  char*          stack;
  int            stacksize;
#endif

  int            num_ports;  /* neel */
  CommPort**     comm_stuff; /* neel */
  PCMutex*       comm_lock;
  int            num_rcvd_messages;
  MsgInfo*       msg_info; /* neel ? required ? */
  MsgInfo*       expected_msg_info; /* neel */ /* only 2 fields used */

  int            registry;   /* neel */
  void*          ret_val;    /* return value from the thread */

#ifdef _JOIN_BY_COND
  PCMutex*       join_lock;
  PCCondition*   join_cond;
#endif
  
  Boolean        cancelled;

  int            sticky;  /* whether the thread can be moved or not */
  int            runs_on; /* on which the thread is running */
  int            created_on; /* on which the thread is created */
  unsigned long  locks_info; /* local, shared and spinlock */
  int            yield_count; /* temporary: to study shmutex2 & ShMutex*/
#ifdef _STATISTICS_ON
  int            sh_lock_access_count; /* # of times a shared lock access was done */
  int            proc_lock_access_count; /* # of spin lock accesses */
  int            num_yields;
  int            num_blocks;
#endif

} Thread;

/* return version number */
char* newthreadsVersion();

/* threadStart "fires up" a thread.  The thread will start running sometime
 * soon after the call.  The thread will execute the procedure-call func(arg).
 */
void threadStart(Thread* t, ThreadFunction func, void* arg, int sticky, const int reg_id);

/* threadBlock causes the current thread to block indefinitely.  The thread 
 * will not resume until somebody calls threadUnblock on it.  threadUnblock
 * makes a blocked thread runnable again. 
 */
void threadBlock();
void threadUnblock(Thread*);

/* threadYield gives up the CPU so another thread can run.  */
void threadYield();

/* cancel a thread */
int threadCancel(Thread*);
/* don't call threadHandoffCPU unless you know what you're doing */
void threadHandoffCPU(Thread*, ThreadStatus);

/* threadJoin .... wait till a thread is done! */
void threadJoin(Thread*, void**);

/* threadExit causes the current thread to exit.  The thread will never run
 * again, but it can be re-used by calling threadStart on it again.  Note 
 */
void threadExit(void*);


/* threadUsesFloatingPoint tells the scheduler whether or not the current
 * thread is using the floating-point registers.  (If a thread is not using
 * these registers, it can be context-switched more quickly.)  By default,
 * threads are assumed to be using floating-point.  This procedure can be
 * called as often as desired to turn floating-point register-saving on and
 * off.
 */
void threadUsesFloatingPoint(int usesIt);

void nodeExit();


/*************************************************************************
 *
 * communication
 */


void Tsend(void* buf, int size, int portNum, 
	   int threadId, int pc, int nodeNum);  
int  Trecv(void* buf, int maxSize, int portNum, 
	   int thread_id, int pc,  int node_num);
void threadPoll();

int  clearMessages();
void clearOneMessage();
void threadCommBlock(int node, int pc, int thr, int port_num);
void threadCommUnblock(Thread* t);

int match_msg_fields(MsgInfo* expected_msg_info, int port_num, int thread_id,
		     int pc, int node_num);


/************Thread Private Data - Added by Neel(4/13/94)*****************/
int thread_keycreate(thread_key_t* key);
int thread_setspecific(Thread* tptr, thread_key_t key, thread_data_t value);
int thread_getspecific(Thread* tptr, thread_key_t key, thread_data_t* value);

/************Thread Registry - Added by Neel(4/15/94)*********************/
/* Thread registry-related data structures */
typedef struct _RegistryList 
{
  Thread* thread_ptr;
  int     thread_id;
  struct  _RegistryList* next;
} RegistryList;

void registerThread(Thread* thread_ptr, const int pre_registry);
void unregisterThread(Thread* thread_ptr);
int NumRegisteredThreads();
void  PrintRegistry();
Thread* ThreadLookup(int thread_id);
int ThreadGetRegId(Thread* tptr);

int is_pending_messages(Thread* t);
int get_num_pending_messages(Thread* t);
int set_num_pending_messages(Thread* t, const int val);
int incr_num_pending_messages(Thread* t);
int decr_num_pending_messages(Thread* t);

/* type of the asynchronous message handlers... */
typedef void (*AsyncMsgHandlerType)(const void* const buf, const int size); 

/*********** Context-level information ***********************************/

typedef struct _Context
{
   int contextId;
   int numProcessors;                         /* number of processors in the 
				  	         processor context */
   Thread* readyQhead_ptr;
   Thread* readyQtail_ptr;
   /* one may use spinlocks or multi-threaded locks */
   PCSpinlock* readyQlock_ptr;

   int registry_index;
   int num_registered_threads;
   RegistryList* registry_list;
   PCSpinlock* registry_lock;

#ifdef _SINGLE_PROCESSOR_PER_PC
   Thread* thisThread_ptr;
   Thread* mainThread_ptr;
   Thread* yielder; 
#else
   Thread* thisThreadList[MAX_PROCESSORS];         /* current thread */
   Thread* mainThreadList[MAX_PROCESSORS];         /* the main sticky */
						   /* thread created, */
						   /* on each */
						   /* processor, which */
						   /* runs the other */
						   /* threads */  
   Thread* yielder[MAX_PROCESSORS];
#endif

#ifndef _SINGLE_PC_PER_SD
   Thread* watchdogThread;
#endif

   int     num_yielders;

   MemoryHandle sharedMem; /* assuming only one processor context */
   PCProcBarrier barrier_p;   /* barrier for the processor context */
   PCThrBarrier* main_thr_barrier_p; /* thread barrier for just the */
				    /* main threads */  

  /* communication related per-context information */
   PCMutex* theWaitersLock; 
   CommWaiters** theWaiters; /* filled in my initCommPorts */
   int num_recv_waiters;  /* number of threads waiting for a message */
			  /* to arrive */  
   int numPorts;        /* number of types of messages that can be passed */

   struct BufferDescriptor* freeBufDescs; /* should have a lock around it */
   PCSpinlock* freeBufDescsLock;

#ifdef _SINGLE_PROCESSOR_PER_PC
   Thread* oldThread;
   ThreadStatus threadState;
#else
  /* context switch stack stuff: used during context switches */
   Thread* oldThreads[MAX_PROCESSORS]; /* an array of pointers of size # processors */
   ThreadStatus threadStates[MAX_PROCESSORS]; /* an array of status of size # processors */
#endif

#ifndef _SINGLE_PC_PER_SD
  /* mail box.. for pc -> pc message passing */
   MailBox* mailBox;
#endif

#ifdef _COMMUNICATION_ON
  AsyncMsgHandlerType msgHandlers[NumMsgPorts];
#endif
   
} Context;

#ifdef _COMMUNICATION_ON
void ll_add_async_msg_handler(const int msg_id,
			      AsyncMsgHandlerType handler);
AsyncMsgHandlerType ll_get_async_msg_handler(const int msg_id);
#endif

typedef struct
{
  MemoryHandle mem;
  SDProcBarrier proc_barrier_p;
  SDShallowThrBarrier* main_thr_barrier_p;
#ifdef _SINGLE_PC_PER_SD
  Context onlyContext;
#else
  Context contextList[MAX_CONTEXTS];
#endif
  int numContexts;
  int numProcessors;
  int  neel_pdata_key_index;
#ifndef _SINGLE_SD_PER_D
  Thread* msgThread;
#endif
#ifndef _SINGLE_PC_PER_SD
MailEntry* freeMailEntries;
SDSpinlock* freeMailEntriesLock;
#endif

} SubDomainCDataStruct;

extern SubDomainCDataStruct subdomainData;

typedef struct
{
  int myNodeNum;
  int numNodes;
  DomShallowThrBarrier* main_thr_barrier_p;
  MemoryHandle mem;
} DomainCDataStruct;
extern DomainCDataStruct domainData;
   

/********** 2 levels of initialization ***********************************/
void InitPCPackage();  
void CreateContexts(const int num_contexts, const int num_procs_per_context);
void FireupContexts();


/* thread space allocation / freeing */
Thread* allocate_thread(MemoryHandle mem, const int stack_size);
void free_thread(MemoryHandle  mem, Thread* tptr);

/********** memory allocation stuff **************************************/

/* memory routines for a particular pc access */
void* shared_malloc(MemoryHandle mem, const int size);
void shared_free(MemoryHandle mem, void* ptr);

/* memory routines for across pc access */
void* global_shared_malloc(const int size);
void global_shared_free(void* ptr);

/*************************************************************************/

/*************************************************************************/
 
/****** added to permit thread->thread and rope->rope message passing*****
 Assuming a message type is a signed integer
 The lower 2 bytes (16 bits) are used for actual message types 
   thus, a message type can be between 0 -> 2^16 - 1 (65535)
 The higher 2 bytes are used for encoding the thread/rope ids
 The highest bit is not used (because signed integer) 
 If the second highest bit is set it is a rope else it is a thread
 A thread/rope id is the remaining high 14 bits 
   Thus a thread/rope id can be between 0 -> 2^14 - 1 (16383)
*************************************************************************/

/* this should be the thread_id on the 0th procesor of the PC */
#define MAIN_THR 1 
/* this should be the MSG_THR on any PC */
#define MSG_THR 9999
#define ANY_THR -1
#define ALL_THR -2
#define ANY_PC  -1
#define ALL_PC -2
#define ANY_SD -1
#define ALL_SD -2
#define ANY_MSG -1

#define IsRope(x)   ((int)(x) & 0x10000000)
#define RopeId(x)   (((int)(x) & 0x0fff0000) >> 16)
#define ThreadId(x) (((int)(x) & 0x0fff0000) >> 16)
#define MsgType(x)  ((int)(x) & 0x0000ffff)
#define ANY_THREAD_CODE 0xfff
#define BuildMsgType(pnum,tid) (((tid)==ANY_THR) \
				? ((ANY_THREAD_CODE << 16) |  (int)(pnum))\
				: (((int)(tid) << 16) | (int)(pnum)))
#define IsMsgTypeInRange(x) (!((x) & 0xffff0000))
#define IsThrOrRope(x) (!((x) & 0xffff0000))
#define IsRopeIdInRange(x) (!((x) & 0xfffff000))
#define IsThreadIdInRange(x) (!((x) & 0xfffff000))
#define MarkRope(x) ((x)|0x1000)


#if 0
#ifndef _PARAGON
#ifdef assert
#undef assert
#define assert(x) {char buf[32]; \
		     if (!(x))\
		     {\
		        sprintf(buf,"[%d:%d:%d]:%s", ll_CurrentContextId(), \
				ll_MyProcessorNum(), \
				ThreadGetRegId(ll_ThisThread()),#x); \
			__assert(buf, __FILE__, __LINE__);\
	             }\
     		   }
#endif
#endif
#endif

#include "inlines.h"

#endif /*  _New_Threads_H */

