/***************************************************************************
 * First Cut at Ports0 interface: 
 * Author: Neel Sundaresan, Indiana University
 ***************************************************************************/

#ifndef PORTS0_H
#define PORTS0_H
#include <sys/stat.h>
#include <stdlib.h>
#include <assert.h>
#include "neelthreads.h"


/*--------------------------------------------------------------------------*
 * data types
 *--------------------------------------------------------------------------*/
typedef Thread* ports0_thread_t;
typedef size_t ports0_threadattr_t; /* at this point only size is supported */
typedef void* ports0_thread_once_t;
typedef thread_key_t ports0_thread_key_t;
typedef PCMutex* ports0_mutex_t;
typedef void* ports0_mutexattr_t;
typedef PCCondition* ports0_cond_t;
typedef void* ports0_condattr_t;


extern int ports0_init(int argc, char**argv, char*package_id);
extern int ports0_shutdown(void);


/* -- thread attributes -- */

extern int ports0_threadattr_init(ports0_threadattr_t *attr);

extern int ports0_threadattr_destroy (ports0_threadattr_t *attr);

extern int ports0_threadattr_setstacksize (ports0_threadattr_t *attr,
                                           size_t stacksize);

extern int ports0_threadattr_getstacksize (ports0_threadattr_t *attr,
                                           size_t *stacksize);

/* -- thread management -- */

/* caveat:(Neel:=  the start function was void*(*)(void*), changed it
  to void (*)(void*).) note: no big deal changing it back! */
extern int ports0_thread_create (ports0_thread_t *thread,
                                 const ports0_threadattr_t *attr,
                                 void (*start)(void*),
                                 void *arg);

extern int ports0_thread_once (ports0_thread_once_t *once_control,
                               void (*start)(void));

extern int ports0_thread_exit (void *value);

extern int ports0_thread_yield (void);

extern ports0_thread_t ports0_thread_self (void);
extern int ports0_thread_equal (ports0_thread_t t1,
                                ports0_thread_t t2);

/* -- thread-specific data -- */

extern int ports0_thread_key_create (ports0_thread_key_t *key,
                                     void (*destructor)(void*));

extern int ports0_thread_key_delete (ports0_thread_key_t key);

extern int ports0_thread_setspecific (ports0_thread_key_t key,
                                      const void *value);

extern void *ports0_thread_getspecific (ports0_thread_key_t key);

/* -- mutexes -- */

extern int ports0_mutexattr_init (ports0_mutexattr_t *attr);

extern int ports0_mutexattr_destroy (ports0_mutexattr_t *attr);

extern int ports0_mutex_init (ports0_mutex_t *mutex,
                              ports0_mutexattr_t *attr);

extern int ports0_mutex_destroy (ports0_mutex_t *mutex);

extern int ports0_mutex_lock (ports0_mutex_t *mutex);

extern int ports0_mutex_trylock (ports0_mutex_t *mutex);

extern int ports0_mutex_unlock (ports0_mutex_t *mutex);

/* -- condition variables -- */

extern int ports0_condattr_init (ports0_condattr_t *attr);

extern int ports0_condattr_destroy (ports0_condattr_t *attr);

extern int ports0_cond_init (ports0_cond_t *cond,
                             ports0_condattr_t *attr);

extern int ports0_cond_destroy (ports0_cond_t *cond);

extern int ports0_cond_broadcast (ports0_cond_t *cond);

extern int ports0_cond_signal (ports0_cond_t *cond);

extern int ports0_cond_wait (ports0_cond_t *cond,
                             ports0_mutex_t *mutex);

/* -- reentrant functions -- */

extern int ports0_reentrant_lock (void);

extern int ports0_reentrant_unlock (void);

extern void *ports0_malloc (size_t bytes);

extern void *ports0_realloc (void *p,
                             size_t bytes);

extern void *ports0_calloc (size_t nobj,
                            size_t bytes);

extern int ports0_free (void *ptr);

extern int ports0_open (char *path,
                        int flags,
                        int mode);

extern int ports0_close (int fd);

extern int ports0_read (int fd,
                        char *buf,
                        int nbytes);

extern int ports0_write (int fd,
                         char *buf,
                         int nbytes);

extern int ports0_fstat (int fd,
                         struct stat *buf);

extern int ports0_lseek (int fd,
                         off_t offset,
                         int whence);


/* the following was added to ports0 by Neel for his selfish satisfaction */
int ports0_thread_join(ports0_thread_t* tp, void** ret_val_p);

#endif  PORTS0_H
